import numpy as np
import random
import bisect
from itertools import product

MaxArcCount = 2  # 16;
ChainLength = 2
IterCountForBuild = 1000000  # 6000000 10000000 15000000;
StartPhero = 0.1
AntPhero = 0.1
AntiPhero = 0.005
AntStepCount = 3

IntervalCount = 40
IntervalStep = 0.1


def unnormalize(series, recovering_values):
    series = ((recovering_values[0] + recovering_values[1]) + series * (recovering_values[1] - recovering_values[0])) / 2.0
    return series

def zero_sequence_generator(length, max_step):
    return list(product(range(0, max_step + 1), repeat=length))

def select_next_restoring_step(current_sequence, current_steps, relation_tensor, max_step,
                               interval_number=IntervalCount, dimensions = 1):
    options = []
    probabilities = []
    possible_sequence = np.zeros((len(current_sequence) + 1, dimensions))
    possible_sequence[:len(current_sequence), :] = current_sequence

    possible_steps = np.zeros(len(current_steps) + 1)
    possible_steps[:len(current_steps)] = current_steps

    for i in range(max_step):
        for interval_seq in zero_sequence_generator(dimensions, interval_number - 1):
            current_sequence[-1, :] = interval_seq
            possible_sequence[-1, :] = i
            key = str([current_sequence, current_steps])
            if key in relation_tensor:
                options.append([interval_seq, i])
                probabilities.append(relation_tensor[key])
    if not options:
        return -1, 0

    probabilities = np.array(probabilities) / np.sum(probabilities)
    option_number = np.random.choice(len(options), p=probabilities)
    result = options[option_number]
    return result[0], result[1]

def interval_choose(value, intervals):
    return [bisect.bisect_left(intervals[i], value[i]) for i in range(len(value))]

def restore_series(relation_tensor, average_interval_values, recovering_values, intervals, start_values, length=10000,
                   max_step=MaxArcCount, vector_len=ChainLength, max_step_number=1000, max_iter=100000, dimensions = 1):
    series = np.zeros((length, dimensions))
    series[:len(start_values),:] = start_values
    current_elem_counts = np.zeros(len(series))
    step_counter = 0

    for i in range(max_step_number):
        if not step_counter % 100:
            print(step_counter // 100, '% done', sep='')

        step_counter += 1
        current_position = random.randrange(5)
        while np.isnan(series[current_position]).any():
            current_position = random.randrange(5)

        current_sequence = np.zeros((vector_len, dimensions))
        current_steps = np.array(np.random.randint(1, max_step, size=vector_len - 1))

        current_sequence[0] = interval_choose(series[current_position], intervals)
        for j in range(vector_len - 1):
            current_position += current_steps[j]
            current_sequence[j + 1] = interval_choose(series[current_position], intervals)

        for iteration in range(max_iter):
            if current_position >= length:
                break
            next_sequence, next_step = select_next_restoring_step(current_sequence, current_steps, relation_tensor,
                                                                  max_step, len(intervals), len(current_sequence[0]))
            if next_sequence == -1:
                break
            current_position += next_step
            current_elem_counts[current_position] += 1

            if np.isnan(series[current_position]).any():
                series[current_position] = average_interval_values[next_sequence, :]
            else:
                series[current_position] = (series[current_position]* (current_elem_counts[current_position] - 1) + average_interval_values[next_step])/\
                                           current_elem_counts[current_position]

            current_sequence = np.append(current_sequence[1:], next_sequence, axis= 0)
            current_steps = np.append(current_steps[1:], next_step)

    return unnormalize(series, recovering_values)

